#! /usr/bin/env python3
###############################################################################
#
# Copyright 2024
# ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

###############################################################################

# Author: Eric Kooistra
# Purpose: FIR filter functions for DSP
# Description:
#
# References:
# [1] dsp_study_erko.txt
# [2] https://github.com/chipmuenk/dsp/blob/main/notebooks/02_LTF/LTF-IIR-allgemein.ipynb

import numpy as np
from scipy import signal
from .utilities import c_rtol, c_atol, ceil_div, ceil_pow2, is_even, is_symmetrical
from .fourier import fourier_interpolate, dtft, estimate_gain_at_frequency


###############################################################################
# Windowed sinc filter design
###############################################################################

def raised_cosine_response(Ntaps, Nsps, roBeta):
    """Generate a raised cosine (RC) FIR filter impulse response.

    Input:
    . Ntaps : FIR filter length
    . Nsps: symbol period Tsymbol in number of samples per symbol
    . roBeta : Roll off factor in [0, 1.0], BW = (1 + roBeta) / Tsymbol, so:
       - roBeta = 0.0: rectangular spectrum with BW = 1 / Tsymbol
       - roBeta = 1.0: cosine spectrum with BW = 2 / Tsymbol
    Return:
    . hRc : impulse response of the raised cosine filter.
    """
    # time axis
    Tsymbol = Nsps  # normalized sample period Ts = 1.0
    tIndices = np.arange(Ntaps)
    tCenter = (Ntaps - 1) / 2
    t = tIndices - tCenter

    # sinc term, can use array assignment because sinc(1 / 0) = 1
    hRc = 1 / Tsymbol * np.sinc(t / Tsymbol)  # np.sinc(x) = sin(pi x) / (pi x)

    # apply cos term, use for loop instead of array assignment, to detect divide by 0
    for tI in tIndices:
        t = tI - tCenter
        if np.abs(t) != Tsymbol / (2 * roBeta):
            hRc[tI] *= np.cos(np.pi * roBeta * t / Tsymbol) / (1 - (2 * roBeta * t / Tsymbol)**2)
    return hRc


def square_root_raised_cosine_response(Ntaps, Nsps, roBeta):
    """Generate a square root raised cosine (SRRC) FIR filter impulse response.

    Reference:
    . [HARRIS section 4.3]
    . https://en.wikipedia.org/wiki/Root-raised-cosine_filter

    Input:
    . Ntaps : FIR filter length
    . Nsps: symbol period Tsymbol in number of samples per symbol
    . roBeta : Roll off factor in [0, 1.0]
       - roBeta = 0.0: rectangular spectrum with BW = 1 / Tsymbol
       - roBeta = 1.0: sqrt cosine spectrum with BW = 2 / Tsymbol
    Return:
    . hSrRc : impulse response of the square root raised cosine filter.
    """
    # time axis
    Tsymbol = Nsps  # normalized sample period Ts = 1.0
    tIndices = np.arange(Ntaps)
    tCenter = (Ntaps - 1) / 2
    t = tIndices - tCenter

    # numerator term, using array assignment
    hSrRc = 1 / Tsymbol * (np.cos(np.pi * (1 + roBeta) * t / Tsymbol) * 4 * roBeta * t / Tsymbol +
                           np.sin(np.pi * (1 - roBeta) * t / Tsymbol))

    # apply denumerator term, use for loop instead of array assignment, to detect divide by 0
    for tI in tIndices:
        t = tI - tCenter
        if t == 0.0:
            hSrRc[tI] = 1 / Tsymbol * (1 + roBeta * (4 / np.pi - 1))
        elif np.abs(t) == Tsymbol / (4 * roBeta):
            hSrRc[tI] = 1 / Tsymbol * roBeta / np.sqrt(2) * \
                ((1 + 2 / np.pi) * np.sin(np.pi / (4 * roBeta)) +
                 (1 - 2 / np.pi) * np.cos(np.pi / (4 * roBeta)))
        else:
            hSrRc[tI] /= (1 - (4 * roBeta * t / Tsymbol)**2) * (np.pi * t / Tsymbol)
    return hSrRc


###############################################################################
# FIR Filter design
###############################################################################

def nof_taps_kaiser_window(fs, fpass, fstop, atten_db):
    """Number of FIR LPF taps using Kaiser window based design

    Reference: [HARRIS 3.2, Fig. 3.8 for kaiserBeta]
    """
    df = fstop - fpass
    return int((fs / df) * (atten_db - 8) / 14)


def nof_taps_remez(fs, fpass, fstop, atten_db):
    """Number of FIR LPF taps using remez = Parks-McClellan based design.

    Reference: [HARRIS 3.3, LYONS 5.6]
    """
    df = fstop - fpass
    return int((fs / df) * (atten_db / 22))


def ideal_low_pass_filter(Npoints, Npass, bandEdgeGain=1.0):
    """Derive FIR coefficients for prototype low pass filter using ifft of
    magnitude frequency response.

    The Npoints defines the double sided spectrum, so Npass = 2 * fpass / fs
    * Npoints, where fpass is the positive cutoff frequency of the LPF.

    Input:
    - Npoints: Number of points of the DFT in the filterbank
    - Npass: Number of points with gain > 0 in pass band.
    - bandEdgeGain : Gain at band edge
    Return:
    - h: FIR coefficients from impulse response
    - f: normalized frequency axis for HF, fs = 1
    - HF: frequency transfer function of h
    """
    # Magnitude frequency reponse
    HF = np.zeros([Npoints])
    HF[0] = bandEdgeGain
    HF[1 : Npass - 1] = 1
    HF[Npass - 1] = bandEdgeGain
    # Zero center HF to make it even
    HF = np.roll(HF, -(Npass // 2))
    f = np.arange(0, 1, 1 / Npoints)
    # Filter impulse response
    h = np.fft.ifft(HF).real  # imag is 0 for even HF
    h = np.roll(h, Npoints // 2)
    return h, f, HF


def design_fir_low_pass_filter(method,
                               Ncoefs, fpass, fstop, fcutoff=0, cutoffGain=0.5, rippleWeights=[1, 1],
                               kaiserBeta=0, fs=1.0,
                               cutoffBW=0, verbosity=1):
    """Derive FIR coefficients for low pass filter

    Use method 'firls' or 'remez', fs = 1.0

    Derive symmetrical FIR coeffients for linear phase. For symmetrical
    coefficient FIR filters the group delay is (Ncoefs - 1) / 2, so
    choose Ncoefs is odd to have integers number of samples output delay.

    Normalize DC gain to 1.0.

    Remark:
    . The actual LPF gain at fcutoff will differ from specified cutoffGain by about 1%.

    Workaround:
    . Method 'firls' does not support zero BW at fcutoff. With single fcutoff point error occurs: 'bands must be
      monotonically nondecreasing and have width > 0'. Work around: use a very narrow band cutoffBW at fcutoff.
      The cutoffBW band has to be much smaller than the transition band transistionBW, to avoid a flattened gain
      around fcutoff, but also not too small, to because then the cutoffGain at fcutoff is not achieved.
      Default force cutoffBW = transistionBW / 10 for 'firls' when input cutoffBW=0. For 'remez' method the
      cutoffBW=0 effectively, without work around, because 'remez' does support single fcutoff value specification.

    Input:
    - method: FIR filter design method 'firls' or 'remez'
    - Ncoefs: Actual number of prototype FIR filter coefficients, any extra
      coefs are zero
    - fpass: pass band frequency
    - fstop: stop band frequency
    - fs: sample frequency, 0 < fpass < fcutoff < fstop < fNyquist = fs / 2
    - fcutoff: when fcutoff > 0, then define cutoff frequency point in transition band, fpass < fcutoff < fstop
    - cutoffGain: normalized LPF gain at fcutoff
    - rippleWeights: relative ripple factors for pass band, optional fcutoff, and stop band
    - kaiserBeta: When kaiserBeta > 0, then additionally apply a Kaiser window on FIR
      coefficients
    - cutoffBW: workaround bandwidth at fcutoff for firls, because firls does not support zero BW at fcutoff.
    - verbosity: when > 0 print() status, else no print()
    Return:
    - h: FIR coefficients for requested Ncoefs
    """
    fNyquist = fs / 2
    lpBW = fpass + fstop  # low pass bandwidth is twice center of transition band
    transistionBW = fstop - fpass

    # Initial FIR filter length N
    # . Use interpolation of factor Q shorter filter to ensure the FIR filter design converges
    #   and to speed up calculation. N >> NcoefsFirlsMax = 1024 is not feasible.
    # . The passband ripple and stopband attenuation depend on the transition bandwidth
    #   and the rippleWeights. Choose transition band width ~< fpass, to ensure the FIR filter
    #   design converges and improve the passband ripple and stopband attenuation. A too large
    #   transition bandwidth also gives the design too much freedom and causes artifacts in
    #   the transition band.
    if method == 'firls':
        NcoefsFirlsMax = 1024
        Q = ceil_div(Ncoefs, NcoefsFirlsMax)
        # Band edges should be less than fNyquist = fs / 2
        if fstop * Q >= fNyquist:
            Q = int(fNyquist / fstop)
        N = Ncoefs // Q
        # If necessary -1 to make odd, because firls only supports odd number of FIR coefficients
        if is_even(N):
            N = N - 1
    if method == 'remez':
        NcoefsRemezMax = 512
        Q = ceil_div(Ncoefs, NcoefsRemezMax)
        # Band edges should be less than fNyquist = fs / 2
        if fstop * Q >= fNyquist:
            Q = int(fNyquist / fstop)
        N = Ncoefs // Q

    # Low pass transition band
    f_pb = fpass * Q  # pass band frequency
    f_co = fcutoff * Q  # optional cutoff frequency in transition band
    f_sb = fstop * Q  # stop band frequency
    f_tb = transistionBW * Q

    # Calculate initial FIR filter for length N
    if method == 'firls':
        if not fcutoff:
            hFir = signal.firls(N, [0, f_pb, f_sb, fNyquist],
                                   [1, 1, 0, 0], rippleWeights, fs=fs)
        else:
            if cutoffBW == 0:
                b_co = f_tb / 10  # default
            else:
                b_co = cutoffBW * Q  # from input argument
            hFir = signal.firls(N, [0, f_pb, f_co - b_co / 2, f_co + b_co / 2, f_sb, fNyquist],
                                   [1, 1, cutoffGain, cutoffGain, 0, 0], rippleWeights, fs=fs)
    if method == 'remez':
        if not fcutoff:
            hFir = signal.remez(N, [0, f_pb, f_sb, fNyquist],
                                   [1, 0], rippleWeights, fs=fs)
        else:
            hFir = signal.remez(N, [0, f_pb, f_co, f_co, f_sb, fNyquist],
                                   [1, cutoffGain, 0], rippleWeights, fs=fs)

    # Additionally apply a Kaiser window, with kaiserBeta = 1 like in pfs_coeff_final.m, this improves
    # the stopband attenuation near the transition band somewhat
    # . kaiserBeta: 0 rect, 5 hamming, 6 hanning
    if kaiserBeta:
        win = signal.windows.kaiser(N, kaiserBeta)
        hFir *= win

    # Normalize DC gain
    dcSum = np.sum(hFir)
    hFir *= 1.0 / dcSum

    if len(hFir) == Ncoefs:
        h = hFir
    else:
        # Use Fourier interpolation to create final FIR filter coefs when
        # Q > 1 or when N != Ncoefs
        HFfir = np.fft.fft(hFir)
        h = fourier_interpolate(HFfir, Ncoefs, verbosity=0)

    if verbosity:
        print('> design_fir_low_pass_filter():')
        print('. Method              = %s' % method)
        print('. Q                   = %f' % Q)
        print('. fpass               = %f' % fpass)
        print('. fstop               = %f' % fstop)
        print('. lpBW                = %f' % lpBW)
        print('. transistionBW       = %f' % transistionBW)
        if fcutoff:
            print('. fcutoff             = %f' % fcutoff)
            print('. cutoffGain          = %f' % cutoffGain)
        print('. fNyquist            = %f' % fNyquist)
        print('. fs                  = %f' % fs)
        print('. Ncoefs              = %d' % len(h))
        print('. DC sum              = %f' % np.sum(h))
        print('. Symmetrical coefs   = %s' % is_symmetrical(h))
        print('')
    return h


def design_fir_low_pass_filter_adjust(method,
                                      Ncoefs, fpass, fstop, fcutoff, cutoffGain=0.5, rippleWeights=[1, 1],
                                      kaiserBeta=0, fs=1.0, resolution=1e-4, verbosity=1):
    """Derive FIR coefficients for low pass filter for exact fcutoff

    Uses design_fir_low_pass_filter() but achieves cutoffGain at fcutoff by adjusting fpass or fstop, and
    fpass < fcutoff < fstop.

    Input:
    . resolution: Resolution measure for frequency and for gain. It appears that at some fine resolution
      even finer resolution does not yield more accurate cutoffGain. Possibly due to the method.
    Return:
    . h: LPF impulse response
    """
    # Determine whether to adjust fpass or fstop
    nofIterations = 1
    nofIterationsMax = 100
    Ndtft = ceil_pow2(max(Ncoefs, 1 / resolution) * 10)
    h = design_fir_low_pass_filter(method, Ncoefs, fpass, fstop, rippleWeights=rippleWeights,
                                   kaiserBeta=kaiserBeta, fs=fs, verbosity=0)
    _, fn, HF = dtft(h, Ndtft)
    f = fn * fs
    fIndex, fValue, fGain = estimate_gain_at_frequency(f, HF, fcutoff)
    if fGain < cutoffGain:
        adjustBand = 0  # increase fpass
        adjustBW = fcutoff - fpass
    else:
        adjustBand = 1  # decrease fstop
        adjustBW = fstop - fcutoff

    # Iterate to achieve cutoffGain at fcutoff
    fRadix = 8
    fStep = adjustBW / fRadix
    gainResolution = cutoffGain * resolution
    freqResolution = 0.1 * fs / Ndtft
    FP = fpass
    FS = fstop
    result = False
    while True:
        # Calculate iteration
        nofIterations += 1
        h = design_fir_low_pass_filter(method, Ncoefs, FP, FS, rippleWeights=rippleWeights,
                                       kaiserBeta=kaiserBeta, fs=fs, verbosity=0)
        _, fn, HF = dtft(h, Ndtft)
        fIndex, fValue, fGain = estimate_gain_at_frequency(f, HF, fcutoff)
        # print('DEBUG: FP, FS, fIndex, fValue, fGain = %s' % str((FP, FS, fIndex, fValue, fGain)))
        if np.abs(fGain - cutoffGain) < gainResolution:
            result = True
            break
        if fStep < freqResolution:
            result = True
            break
        if nofIterations > nofIterationsMax:
            print('ERROR: Too many iterations.')
            break
        # Prepare next iteration
        if adjustBand == 0:
            if fGain < cutoffGain:
                # Keep on increasing FP
                FP += fStep
                # print('DEBUG: Increase FP = %s' % str(FP))
                # If necessary increase FS to keep fcutoff, this can occur when rippleWeights for stop band >
                # rippleWeights for pass band
                if FP >= fcutoff:
                    FP = fpass
                    FS += fStep
                    # print('DEBUG: Need to increment FS = %s' % str(FS))
            else:
                # FP too high, one fStep back and increment FP with reduced fStep
                FP -= fStep
                fStep /= fRadix
                FP += fStep
                # print('DEBUG: FP too high, decrease fStep: FP, fStep = %s' % str((FP, fStep)))
        else:
            if fGain > cutoffGain:
                # Keep on decreasing FS
                FS -= fStep
                # print('DEBUG: Decrease FS = %s' % str(FS))
            else:
                # FS too low, one fStep back and decrement FS with reduced fStep
                FS += fStep
                fStep /= fRadix
                FS -= fStep
                # print('DEBUG: FS too low, decrease fStep: FS, fStep = %s' % str((FS, fStep)))

    # Return result
    if result:
        if verbosity:
            # Repeat design_fir_low_pass_filter() for logging
            h = design_fir_low_pass_filter(method, Ncoefs, FP, FS, rippleWeights=rippleWeights,
                                           kaiserBeta=kaiserBeta, fs=fs, verbosity=verbosity)
            print('> design_fir_low_pass_filter_adjust():')
            print('. nofIterations        = %d' % nofIterations)
            print('. FP / fpass           = %f' % (FP / fpass))
            print('. FS / fstop           = %f' % (FS / fstop))
            print('. fcutoff              = %f' % fcutoff)
            print('. fGain                = %f' % fGain)
            print('. fGain**2             = %f' % (fGain**2))
            print('')
    else:
        print('ERROR: Return iteration %d.' % nofIterations)
    return h


def prototype_fir_low_pass_filter(method='firls',
                                  Npoints=1024, Ntaps=16, Ncoefs=1024*16,
                                  hpFactor=0.9, transitionFactor=0.4, stopRippleFactor=1000000, kaiserBeta=0, fs=1.0):
    """Derive FIR coefficients for prototype low pass filter

    Use method 'firls' or 'remez'.

    Default use LPF specification for LOFAR subband filter. For subband filter
    BWbin = fs / Npoints and the number of subbands is Nsub = Npoints / 2.

    For multirate upsampling or or downsampling filter the Npoints is the
    multirate factor Nup in upsampling or Ndown in downsampling. The fpass =
    BWbin / 2 = fNyquist / Npoints, where fNyquist = fs / 2.

    Use hpFactor < 1.0 to increase attentuation at fpass, to reduce disturbance
    due to aliasing.

    The number of prototype FIR filter coefficients Ncoefs <= Npoints * Ntaps.
    For symmetrical-coefficient FIR filters the group delay is
    (Ncoefs - 1) / 2, so choose Ncoefs is odd to have integers number of
    samples output delay.

    Normalize DC gain to 1.0.

    Input:
    - method: FIR filter design method 'firls' or 'remez'
    - Npoints: Number of points of the DFT in the filterbank, or multirate
      factor Nup in upsampling or Ndown in downsampling
    - Ntaps: Number of taps per polyphase.
    - Ncoefs: Actual number of prototype FIR filter coefficients, any extra
      coefs are zero
    - hpFactor : Half power bandwidth of the filter relative to BWbin
    - transitionFactor: transition bandwidth factor relative to fpass
    - stopRippleFactor: stopband ripple factor relative to pass band ripple
    - kaiserBeta: When kaiserBeta > 0 then additionally apply a Kaiser window on FIR
      coefficients
    - fs: sample frequency, for logging
    Return:
    - h: FIR coefficients for requested Ncoefs
    """
    # LPF specification for subband filter
    BWbin = fs / Npoints  # bandwidth of one bin (is subband frequency)
    # Use half power bandwidth factor to tune half power cutoff frequency of LPF, default 1.0
    BWpass = hpFactor * BWbin
    fpass = BWpass / 2  # pass band frequency of bin at DC: -fpass to +fpass
    transitionBW = transitionFactor * fpass  # transition bandwidth
    fcutoff = 0  # not used, use hpFactor instead
    cutoffGain = 0.5
    fstop = fpass + transitionBW  # stop band frequency
    rippleWeights = [1, stopRippleFactor]

    # Design subband filter
    h = design_fir_low_pass_filter(method, Ncoefs, fpass, fstop, fcutoff, cutoffGain, rippleWeights, kaiserBeta, fs)
    return h


###############################################################################
# Filterbank
###############################################################################

def filterbank_frequency_response(HFprototype, Npoints):
    """Aggregate frequency reponse of filterbank

    Sum of bandpass responses for all Npoints frequency bins in a filterbank.

                N-1
    HFbank(w) = sum HF_k(w), with HF_k(w) = H0(w - 2pi/N k)
                k=0

    and H0(w) is the prototype LPF [PROAKIS Eq. 10.9.2].
    """
    Lprototype = len(HFprototype)  # corresponds to 2 pi
    Lbin = Lprototype // Npoints  # corresponds to 2 pi / N
    HFbank = np.zeros(Lprototype, dtype='cfloat')
    for bi in range(Npoints):
        HFbank += np.roll(HFprototype, bi * Lbin)
    return HFbank


def filterbank_impulse_response(hprototype, Npoints):
    """Impulse reponse of filterbank

    Sum of bandpass impulse responses for all Npoints frequency bins in a
    filterbank.

               N-1
    hbank[n] = sum hprototype[n] * exp(-j * 2 * np.pi / Npoints * k * n)
               k=0
    """
    Lprototype = len(hprototype)
    hbank = np.zeros(Lprototype, dtype='cfloat')
    for k in range(Npoints):
        hbank += hprototype * np.exp(-1.0j * 2 * np.pi / Npoints * k * np.arange(Lprototype))
    # hbank.imag is zero
    if not np.allclose(hbank.imag, np.zeros(Lprototype), rtol=c_rtol, atol=c_atol * Npoints):
        print('ERROR: hbank.imag != 0 (max(abs) = %e)' % np.max(np.abs(hbank.imag)))
    return hbank.real
