#! /usr/bin/env python3
###############################################################################
#
# Copyright 2024
# ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

###############################################################################

# Author: Eric Kooistra
# Purpose: Hilbert transform functions for DSP
# Description:
#
# References:
# [1] dsp_study_erko.txt

import numpy as np
from .utilities import is_even


###############################################################################
# Hilbert transform filter
###############################################################################

def hilbert_response(Ntaps):
    """Calculate impulse response of Hilbert filter truncated to Ntaps
    coefficients.

    h(t) = 1 / (pi t) ( 1 - cos(ws t / 2), l'Hoptial's rule: h(0) = 0

    For n = ...-2, -1, 0, 1, 2, ..., and Ts = 1, and Ntaps is
    . odd: t = n Ts
        ht[n] = 1 / (pi n)) ( 1 - cos(pi n))
              = 0, when n is even
              = 2 / (pi n), when n is odd
    . even: t = (n + 0.5) Ts
        ht(n) = 1 / (pi m)) ( 1 - cos(pi m)), with m = n + 0.5
    """
    Npos = Ntaps // 2
    if is_even(Ntaps):
        ht = np.zeros(Npos)
        for n in range(0, Npos):
            m = n + 0.5
            ht[n] = 1 / (np.pi * m) * (1 - np.cos(np.pi * m))
        ht = np.concatenate((np.flip(-ht), ht))
    else:
        Npos += 1
        ht = np.zeros(Npos)
        for n in range(1, Npos, 2):
            ht[n] = 2 / (np.pi * n)
        ht = np.concatenate((np.flip(-ht[1:]), ht))
    return ht


def hilbert_delay(Ntaps):
    """Delay impulse by (Ntaps - 1) / 2 to align with hilbert_response(Ntaps).

    Analytic signal htComplex = htReal + 1j * htImag where:
    . htReal = hilbert_delay(Ntaps)
    . htImag = hilbert_response(Ntaps)

    Only support integer delay D = (Ntaps - 1) / 2, so Ntaps is odd, then return
    htReal, else return None.
    """
    if is_even(Ntaps):
        return None
    D = (Ntaps - 1) // 2
    htReal = np.zeros(Ntaps)
    htReal[D] = 1.0
    return htReal
