#! /usr/bin/env python3
###############################################################################
#
# Copyright 2024
# ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

###############################################################################

# Author: Eric Kooistra
# Purpose: IIR filter design functions for DSP
# Description:
#
# References:
# [1] dsp_study_erko.txt
# [2] https://webaudio.github.io/Audio-EQ-Cookbook/Audio-EQ-Cookbook.txt

import numpy as np


###############################################################################
# IIR Filter design
###############################################################################

def iir_biquad_butter_coefficients(f0, BW, band='bandstop', fs=1):
    """Calculate b0, b1, b2, a0, a1, a2 for biquad IIR filter section.

    Bilinear transform of Butterworth filters, from [2].

    For single biquad bandpass alternatively use signal.iirpeak()
    For single biquad bandstop alternatively use signal.iirnotch()
    """
    w0 = 2 * np.pi * f0 / fs
    Q = f0 / BW
    si = np.sin(w0)
    cs = np.cos(w0)
    alpha = np.sin(w0) / (2 * Q)
    a1 = -2 * cs
    if band == 'lowpass':
        b0 = 1 - cs
        b = [b0 / 2, b0, b0 / 2]
    elif band == 'highpass':
        b0 = 1 + cs
        b = [b0 / 2, -b0, b0 / 2]
    elif band == 'bandstop':
        b = [1, a1, 1]
    elif band == 'bandpass':
        b = [si / 2, 0, -si / 2]
    a = [1 + alpha, a1, 1 - alpha]
    return b, a
