#! /usr/bin/env python3
###############################################################################
#
# Copyright 2024
# ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

###############################################################################

# Author: Eric Kooistra
# Purpose: Plotting functions for DSP
# Description:
#
# References:
# [1] dsp_study_erko.txt
# [2] https://github.com/chipmuenk/dsp/blob/main/notebooks/02_LTF/LTF-IIR-allgemein.ipynb

import numpy as np
from scipy import signal
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import dsp_fpga_lib  # from [2]
from .utilities import pow_db


def plot_time_response(h, title='', color='r', markers=False):
    """Plot time response (= impulse response, window, FIR filter coefficients).

    Input:
    . h: time response
    . title: plot title
    . color: curve color format character
    . markers: when True plot time sample markers in curve
    """
    if markers:
        plt.plot(h, color + '-', h, color + 'o')
    else:
        plt.plot(h, color + '-')
    if title:
        plt.title(title)
    else:
        plt.title('Time response')
    plt.ylabel('Voltage')
    plt.xlabel('Sample')
    plt.grid(True)


def plot_iir_filter_analysis(b, a, fs=1.0, whole=False, Ntime=100, step=False, log=False, show=[]):
    """Plot and print iir filter analysis results.

    Input:
    . b, a: IIR filter coefficients in same format as for scipy.signal.freqz
    . fs: sample frequency
    . whole: False to have only positive frequencies, so 0 to fs / 2 in
      spectrum plots, True to have 0 to fs.
    . Ntime: number of timesamples for impulse response
    . step: False for impulse response, True for step response
    . show[]: which plots to show from: 'zplane', 'power spectrum',
      'phase spectrum', 'time response'

    Return: z, p, k
    """

    # Plot poles / zeros diagram in z-plane
    if 'zplane' in show:
        fig1, ax1 = plt.subplots(1)
        ax1.set_title('Zeros (o) and poles (x) in z-plane')
        z, p, k = dsp_fpga_lib.zplane(b, a, plt_ax=ax1)  # uses np.roots(a), np.roots(b)
    else:
        z, p, k = dsp_fpga_lib.zplane(b, a)  # no plot, only calculate z, p, k

    if log:
        print('plot_iir_filter_analysis()')
        print('Zeros, poles and gain from b, a coefficients:')
        if len(z) > 0:
            print('. zeros:')
            for zero in z:
                print('  z = %s' % str(zero))
        if len(p) > 0:
            print('. poles:')
            for pole in p:
                print('  p = %s' % str(pole))
        print('. gain: k = %.3f' % k)

        # Derive b, a coefficients back from z, p, k
        print('Coefficients back from z, p, k:')
        print('  b = %s' % str(np.poly(z)))
        print('  a = %s' % str(np.poly(p) / k))
        print('')

    # Plot transfer function H(f), is H(z) for z = exp(j w), so along the unit circle
    # . 0 Hz at 1 + 0j, fs / 4 at 0 + 1j, fNyquist = fs / 2 at -1 + 0j
    # . use whole=False to have only positive frequencies, so 0 to fs / 2
    # . use Nfreq frequency points
    if 'power spectrum' in show or 'phase spectrum' in show:
        Nfreq = 1024
        [f, HF] = signal.freqz(b, a, Nfreq, whole=whole, fs=fs)
        if 'power spectrum' in show:
            fig2, ax2 = plt.subplots(1)
            ax2.plot(f, pow_db(HF))
            ax2.set_title('Power spectrum')
            ax2.set_xlabel('frequency [fs = %f]' % fs)
            ax2.set_ylabel('HF power [dB]')

        if 'phase spectrum' in show:
            fig3, ax3 = plt.subplots(1)
            ax3.plot(f, np.unwrap(np.angle(HF)))
            ax3.set_title('Phase spectrum')
            ax3.set_xlabel('frequency [fs = %f]' % fs)
            ax3.set_ylabel('HF phase [rad]')

    # Plot impulse response
    if 'time response' in show:
        Ts = 1 / fs
        fig4, ax4 = plt.subplots(1)
        # impz uses signal.lfilter(), and cumsum(h) for step response
        [h, t] = dsp_fpga_lib.impz(b, a, FS=fs, N=Ntime, step=step)
        (ml, sl, bl) = ax4.stem(t, h, linefmt='b-', markerfmt='ro', basefmt='k')
        if step:
            ax4.set_title('Step response')
        else:
            ax4.set_title('Impulse response')
        ax4.set_xlabel('time [Ts = %f]' % Ts)
        ax4.set_ylabel('h[n]')
    return z, p, k


def plot_spectra(fn, HF, fs=1.0, fLim=None, dbLim=None):
    """Plot spectra for power, magnitude, phase, real, imag

    Input:
    . fn: normalized frequency axis for HF (fs = 1)
    . HF: spectrum, e.g. frequency transfer function HF = DTFT(h)
    . fs: sample frequency in Hz
    """
    Hmag = np.abs(HF)
    Hphs = np.unwrap(np.angle(HF))
    Hpow_dB = pow_db(HF)  # power response
    f = fn * fs  # scale fn by fs
    flabel = 'Frequency [fs = %f]' % fs

    plt.figure(1)
    plt.plot(f, Hpow_dB)
    plt.title('Power spectrum')
    plt.ylabel('Power [dB]')
    plt.xlabel(flabel)
    if fLim:
        plt.xlim(fLim)
    if dbLim:
        plt.ylim(dbLim)
    plt.grid(True)

    plt.figure(2)
    plt.plot(f, HF.real, 'r')
    plt.plot(f, HF.imag, 'g')
    plt.title('Complex voltage spectrum')
    plt.ylabel('Voltage')
    plt.xlabel(flabel)
    plt.legend(['real', 'imag'])
    if fLim:
        plt.xlim(fLim)
    plt.grid(True)

    plt.figure(3)
    plt.plot(f, Hmag)
    plt.title('Magnitude spectrum')  # = amplitude
    plt.ylabel('Voltage')
    plt.xlabel(flabel)
    if fLim:
        plt.xlim(fLim)
    plt.grid(True)

    plt.figure(4)
    plt.plot(f, Hphs)
    plt.title('Phase spectrum (note -1: pi = -pi)')
    plt.ylabel('Phase [rad]')
    plt.xlabel(flabel)
    if fLim:
        plt.xlim(fLim)
    plt.grid(True)


def plot_power_spectrum(fn, HF, fmt='r', fs=1.0, fLim=None, dbLim=None):
    """Plot power spectrum

    Input:
    . fn: normalized frequency axis for HF (fs = 1)
    . HF: spectrum, e.g. frequency transfer function HF = DTFT(h)
    . fmt: curve format string
    . fs: sample frequency in Hz
    """
    f = fn * fs  # scale fn by fs
    flabel = 'Frequency [fs = %f]' % fs

    plt.plot(f, pow_db(HF), fmt)
    plt.title('Power spectrum')
    plt.ylabel('Power [dB]')
    plt.xlabel(flabel)
    if fLim:
        plt.xlim(fLim)
    if dbLim:
        plt.ylim(dbLim)
    plt.grid(True)


def plot_magnitude_spectrum(fn, HF, fmt='r', fs=1.0, fLim=None, voltLim=None):
    """Plot magnitude (= voltage) spectrum

    Input:
    . fn: normalized frequency axis for HF (fs = 1)
    . HF: spectrum, e.g. frequency transfer function HF = DTFT(h)
    . fmt: curve format string
    . fs: sample frequency in Hz
    """
    f = fn * fs  # scale fn by fs
    flabel = 'Frequency [fs = %f]' % fs

    plt.plot(f, np.abs(HF), fmt)
    plt.title('Magnitude spectrum')
    plt.ylabel('Voltage')
    plt.xlabel(flabel)
    if fLim:
        plt.xlim(fLim)
    if voltLim:
        plt.ylim(voltLim)
    plt.grid(True)


def plot_two_power_spectra(fn1, HF1, name1, fn2, HF2, name2, fs=1.0, fLim=None, dbLim=None, showRoll=False):
    """Plot two power spectra in same plot for comparison

    Input:
    . fn1,fn2: normalized frequency axis for HF1, HF2 (fs = 1)
    . HF1, HF2: spectrum, e.g. frequency transfer function HF = DTFT(h)
    . fs: sample frequency in Hz
    """
    f1 = fn1 * fs  # scale fn1 by fs
    f2 = fn2 * fs  # scale fn1 by fs
    flabel = 'Frequency [fs = %f]' % fs

    if showRoll:
        plt.plot(f1, pow_db(HF1), 'r',
                 f1, np.roll(pow_db(HF1), len(f1) // 2), 'r--',
                 f2, pow_db(HF2), 'b',
                 f2, np.roll(pow_db(HF2), len(f2) // 2), 'b--')
        plt.legend([name1, '', name2, ''])
    else:
        plt.plot(f1, pow_db(HF1), 'r',
                 f2, pow_db(HF2), 'b')
        plt.legend([name1, name2])
    plt.title('Power spectrum')
    plt.ylabel('Power [dB]')
    plt.xlabel(flabel)
    if fLim:
        plt.xlim(fLim)
    if dbLim:
        plt.ylim(dbLim)
    plt.grid(True)


def plot_colorbar(plt):
    # Work around using make_axes_locatable() to avoid:
    # /tmp/ipykernel_9570/3561080291.py:12: MatplotlibDeprecationWarning: Auto-removal of grids by
    # pcolor() and pcolormesh() is deprecated since 3.5 and will be removed two minor releases
    # later; please call grid(False) first.
    # Using plt.grid(False) is not sufficient to avoid the warning.
    ax = plt.gca()
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size='5%', pad=0.1)
    cax.grid(False)
    plt.colorbar(cax=cax)
