#! /usr/bin/env python3
###############################################################################
#
# Copyright 2024
# ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

###############################################################################

# Author: Eric Kooistra
# Purpose: Window functions for DSP
# Description:
#
# References:
# [1] dsp_study_erko.txt

import numpy as np


def raised_cosine_window(N, roBeta):
    """Generate raised cosine shape window, where cosine is from -pi to pi

    Input:
    . N: window length
    . roBeta: Roll off factor in [0, 1.0] as fraction of N
    Return:
    . rcWin: Window with shape of roll up cos, flat top and roll off cos,

    """
    Ncos = int(roBeta * N)
    Nro = Ncos // 2
    rcWin = np.ones(N)
    if Nro > 0:
        roCos = 0.5 + 0.5 * np.cos(2 * np.pi * (np.arange(Ncos) / Ncos - 0.5))
        ro = roCos[0:Nro]
        rcWin[0:Nro] = ro
        rcWin[-Nro:] = np.flip(ro)
    return rcWin


def half_cosine_window(N, roBeta):
    """Generate half cosine shape window, where cosine is from -pi/2 to pi/2

    The roll up cos and roll down cos edges together have roBeta * N points,
    the flat top in the center has (1 - roBeta) * N points. Hence roBeta = 0
    yields rectangular window and roBeta = 1 yields positive half cos window.

    Input:
    . N: window length
    . roBeta: Roll off (= ro) factor in [0, 1.0] as fraction of N
    Return:
    . cWin: Window with shape of roll up cos, flat top and roll off cos.
    """
    Ncos = int(roBeta * N)
    Nro = Ncos // 2
    cWin = np.ones(N)
    if Nro > 0:
        roCos = np.cos(np.pi * (np.arange(Ncos) / Ncos - 0.5))
        ro = roCos[0:Nro]
        cWin[0:Nro] = ro
        cWin[-Nro:] = np.flip(ro)
    return cWin
